//
// Author:  Christoph Dalitz
// Version: 2.0a from 2014/05/23
//
// Computes the symmetry transformation after
// VISAPP 2013 paper by Dalitz, Pohle-Froehlich, and Bolten
// and return the maximum rotational symmetry point
//

#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include "symmetry_transform.h"

#define TRACEPREFIX "trace-"

// command line arguments
char* opt_infile = NULL;
char* opt_mask = NULL;
int opt_r = 0;
double opt_alpha = 0.5;
bool opt_rect = true;
int opt_axial = 3;
int opt_only = 0;
bool opt_log = false;
bool opt_norm = false;
bool opt_trace = false;
int opt_trace_greytrans = 0;
char* opt_outfile = NULL;
const char* usage = "Usage:\n\tsymmetry -r <r> [options] <infile>\nOptions:\n"
  "\t-r <r> maximum radius of symmetry transform\n"
  "\t-alpha <alpha>\n"
  "\t       exponent for normalisation of symmetry score with r^alpha [0.5]\n"
  "\t-square\n"
  "\t       do not try rectangular regions, but only squares (rx=ry)\n"
  "\t-axial <int>\n"
  "\t       criterion for sorting out axial symmetries:\n"
  "\t        0: edge_dir >= 0.27 (VISAPP'13 paper)\n"
  "\t        1: skel_size > 0.46 or edge_dir >= 0.27\n"
  "\t        2: edge_dir >= 0.285 or anti_par > 0.287 (decision tree)\n"
  "\t        3: QDA on all four features (VISAPP'15 paper, default)\n"
  "\t-only (dark|light)\n"
  "\t       only look for symmetries in dark (or light) objects\n"
  "\t-mask <maskimg>\n"
  "\t       mask image in which black indicates background to be ignored\n"
  "\t       beware that the mask is enlarged by the radius =>\n"
  "\t       use mask images only in combination with small r!\n"
  "\t-o <outpng>\n"
  "\t       draw detected symmetry as rectangle in file <outpng>\n"
  "\t-gradlog\n"
  "\t       transform gradient strength as log(1+|G|)\n"
  "\t-trace write trace of interim images with prefix '" TRACEPREFIX "'\n"
  "\t-trace-greytrans\n"
  "\t       same as -trace, specifies the transformation from float to grey:\n"
  "\t          0 = linear transformation over entire range\n"
  "\t          1 = linear transformation over only positive range\n"
  "\t          2 = like 1, but with subsequent histogram equalization\n"
  "\t          3 = equalization over entire float range\n"
  "\t-v     print version and exit\n";


int main(int argc, char** argv)
{
  cv::Mat img, grey, bgr, imgmask, greymask, result, result_rx, result_ry, gradx, grady;
  double edge_dir, anti_par, skel_size, cov_ratio;
  int rx,ry,i;
  cv::Point maxpoint(0,0);

  // parse command line
  for (i=1; i<argc; i++) {
    if (0==strcmp("-r", argv[i])) {
      i++; if (i<argc) opt_r = atoi(argv[i]);
    }
    else if (0==strcmp("-alpha", argv[i])) {
      i++; if (i<argc) opt_alpha = atof(argv[i]);
    }
    else if (0==strcmp("-square", argv[i])) {
      opt_rect = false;
    }
    else if (0==strcmp("-mask", argv[i])) {
      i++; if (i==argc) {
        fputs(usage, stderr);
        return 1;
      }
      opt_mask = argv[i];
    }
    else if (0==strcmp("-only", argv[i])) {
      i++; if (i==argc) {
        fputs(usage, stderr);
        return 1;
      }
      if (0==strcmp("dark", argv[i])) {
        opt_only = 1;
      }
      else if (0==strcmp("light", argv[i])) {
        opt_only = -1;
      }
      else {
        fputs(usage, stderr);
        return 1;
      }
    }
    else if (0==strcmp("-axial", argv[i])) {
      i++; if (i==argc) {
        fputs(usage, stderr);
        return 1;
      }
      else {
        opt_axial = atoi(argv[i]);
      }
    }
    else if (0==strcmp("-o", argv[i])) {
      i++; if (i<argc) opt_outfile = argv[i];
    }
    else if (0==strcmp("-v", argv[i])) {
      printf("%s, compiled with NUM_THREADS=%i\n", VERSION, NUM_THREADS);
      return 0;
    }
    else if (0==strcmp("-gradlog", argv[i])) {
      opt_log = true;
    }
    else if (0==strcmp("-trace", argv[i])) {
      opt_trace = true;
    }
    else if (0==strcmp("-trace-greytrans", argv[i])) {
      i++; if (i<argc) {
        opt_trace_greytrans = atoi(argv[i]);
        opt_trace = true;
      }
    }
    else if (argv[i][0] == '-') {
      fputs(usage, stderr);
      return 1;
    }
    else {
      opt_infile = (char*)argv[i];
    } 
  }
  if (!opt_r || !opt_infile) {
    fputs(usage, stderr);
    return 1;
  }
  if (opt_mask && (0!=access(opt_mask, R_OK))) {
	fprintf(stderr, "Cannot find file '%s'\n", opt_mask);
    fputs(usage, stderr);
    return 1;
  }

  // load image and call symmetry transform
  img = cv::imread(opt_infile);
  if (!img.data) {
    fprintf(stderr, "cannot open '%s'\n", opt_infile);
    return 1;
  }
  cv::cvtColor(img,grey,CV_RGB2GRAY);
  //cv::GaussianBlur(grey,grey,cv::Size(3,3),1);
  if (opt_log) {
    GradientNormalization gradnorm;
    if (opt_rect) {
      symmetry_transform_rect(grey, result, result_rx, result_ry, opt_r, opt_alpha, opt_only, &gradnorm);
    } else {
      symmetry_transform(grey, result, result_rx, opt_r, opt_alpha, opt_only, &gradnorm);
    }
  } else {
    if (opt_rect) {
      symmetry_transform_rect(grey, result, result_rx, result_ry, opt_r, opt_alpha, opt_only);
    } else {
      symmetry_transform(grey, result, result_rx, opt_r, opt_alpha, opt_only);
    }
  }

  // reset all points that touch mask image to zero
  if (opt_mask) {
    imgmask = cv::imread(opt_mask);
    if (!imgmask.data) {
      fprintf(stderr, "cannot open '%s'\n", opt_mask);
      return 1;
    }
    if (imgmask.rows != result.rows || imgmask.cols != result.cols) {
      fprintf(stderr, "Mask image '%s' must have same size as input image '%s'\n", opt_mask, opt_infile);
      return 1;
    }
    cv::cvtColor(imgmask,greymask,CV_RGB2GRAY);
    int rx,ry;
    for (int y=0; y<greymask.rows; y++) {
      for (int x=0; x<greymask.cols; x++) {
        if (greymask.at<uchar>(y,x) == 0) {
          result.at<double>(y,x) = 0.0;
        }
        else {
          //ry = rx = result_rx.at<double>(y,x);
          //if (opt_rect) ry = result_ry.at<double>(y,x);
          ry = rx = opt_r;
          for (int yy=std::max(0,y-ry); yy<std::min(result.rows,y+ry+1); yy++) {
            for (int xx=std::max(0,x-rx); xx<std::min(result.cols,x+rx+1); xx++) {
              if (greymask.at<uchar>(yy,xx) == 0) {
                result.at<double>(y,x) = 0.0;
                goto endofloop;
              }
            }
          }
        endofloop: ;
        }
      }
    }
  }

  if (opt_trace) {
    float2grey(result, grey, opt_trace_greytrans);
    cv::imwrite(TRACEPREFIX "symmetry.png", grey);
  }

  // find all local maxima and sort them by symmetry score
  // BEWARE: sorting is in ascending order!
  SymmetryPointVector sp;
  if (0 >= points_local_maxima(result, sp)) {
    fputs("no symmetry points found\n", stderr);
    return -2;
  }
  std::sort(sp.begin(), sp.end());

  // find first candidate point that matches rotational criterion
  cv::Sobel(result, gradx, CV_64F, 1, 0);
  cv::Sobel(result, grady, CV_64F, 0, 1);
  if (opt_trace) {
    float2bgr(result, bgr, opt_trace_greytrans);
  }
  for (i=(int)sp.size()-1; i>=0; i--) {
    bool isaxial = false;
    ry = rx = result_rx.at<double>(sp[i].point);
    if (opt_rect) ry = result_ry.at<double>(sp[i].point);

    // classifiction axial versus rotational
    if (opt_axial == 0) {
      edge_dir = edge_directedness_cd(gradx, grady, sp[i].point, 3, 16);
      isaxial = (edge_dir >= 0.27);
    }
    else if (opt_axial == 1) {
      edge_dir = edge_directedness_cd(gradx, grady, sp[i].point, 3, 16);
      PointVector skeleton;
      get_skeleton(sp[i].point, result, &skeleton, 0.5, opt_r);
      skel_size = skeleton_size(sp[i].point, skeleton, rx, ry);
      isaxial = ((skel_size > 0.46) || (edge_dir >= 0.27));
    }
    else if (opt_axial == 2) {
      edge_dir = edge_directedness_cd(gradx, grady, sp[i].point, 3, 16);
      antiparallelity_rpf(gradx, grady, sp[i].point, std::min(rx,ry), anti_par);
      isaxial = ((anti_par > 0.287) || (edge_dir >= 0.285));
    }
    else {
      edge_dir = edge_directedness_cd(gradx, grady, sp[i].point, 3, 16);
      PointVector skeleton;
      get_skeleton(sp[i].point, result, &skeleton, 0.5, opt_r);
      skel_size = skeleton_size(sp[i].point, skeleton, rx, ry);
      antiparallelity_rpf(gradx, grady, sp[i].point, std::min(rx,ry), anti_par);
      cov_ratio = coveigenratio_cd(result, sp[i].point, 3);
      isaxial = axial_qda(edge_dir, skel_size, anti_par, cov_ratio);
    }

    if (!isaxial) {
      maxpoint = sp[i].point;
      if (opt_trace) {
        bgr.at<cv::Vec3b>(sp[i].point) = cv::Vec3b(0,255,0);
        ry = rx = (int)result_rx.at<double>(sp[i].point);
        if (opt_rect) {
          ry = (int)result_ry.at<double>(sp[i].point);
        }
        cv::rectangle(bgr,cv::Point(sp[i].point.x-rx,sp[i].point.y-ry),cv::Point(sp[i].point.x+rx,sp[i].point.y+rx),cv::Scalar(0,255,0));
        //cv::rectangle(bgr,cv::Point(sp[i].point.x-rx-1,sp[i].point.y-ry-1),cv::Point(sp[i].point.x+rx+1,sp[i].point.y+rx+1),cv::Scalar(0,255,0));
      }
      break;
    } else {
      if (opt_trace)
        bgr.at<cv::Vec3b>(sp[i].point) = cv::Vec3b(0,0,255);
        ry = rx = (int)result_rx.at<double>(sp[i].point);
        if (opt_rect) {
          ry = (int)result_ry.at<double>(sp[i].point);
        }
        cv::rectangle(bgr,cv::Point(sp[i].point.x-rx,sp[i].point.y-ry),cv::Point(sp[i].point.x+rx,sp[i].point.y+ry),cv::Scalar(0,0,255));
        //cv::rectangle(bgr,cv::Point(sp[i].point.x-rx-1,sp[i].point.y-ry-1),cv::Point(sp[i].point.x+rx+1,sp[i].point.y+ry+1),cv::Scalar(0,0,255));
    }
  }
  if (opt_trace) {
    cv::imwrite(TRACEPREFIX "localmaxpoints.png", bgr);
  }

  // draw maximum symmetry point in original image
  if (maxpoint.x == 0 && maxpoint.y == 0) {
    fputs("no symmetry point found\n", stderr);
    return 3;
  }
  ry = rx = (int)result_rx.at<double>(maxpoint);
  if (opt_rect) {
    ry = (int)result_ry.at<double>(maxpoint);
  }
  printf("point=(%i,%i) radius=(%i,%i)\n", maxpoint.x, maxpoint.y, rx, ry);
  if (opt_outfile) {
    cv::rectangle(img,cv::Point(maxpoint.x-rx,maxpoint.y-ry),cv::Point(maxpoint.x+rx,maxpoint.y+ry),cv::Scalar(0,0,255));
    cv::imwrite(opt_outfile, img);
  }

  return 0;
}
